# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: base_model.py
@time: 2019-01-19 11:50

"""
import logging
import random
from itertools import chain
from typing import Tuple, Dict

import numpy as np
from keras.preprocessing import sequence
from keras.utils import to_categorical, multi_gpu_model
from sklearn import metrics
from sklearn.utils import class_weight as class_weight_calculte
from sklearn.preprocessing import MultiLabelBinarizer

import kashgari.macros as k
from kashgari.tasks.base import BaseModel
from kashgari.embeddings import BaseEmbedding
from kashgari.type_hints import *
from kashgari.utils.helper import depth_count


class ClassificationModel(BaseModel):

    def __init__(self,
                 embedding: BaseEmbedding = None,
                 hyper_parameters: Dict = None,
                 multi_label: bool = False,
                 **kwargs):
        """

        :param embedding:
        :param hyper_parameters:
        :param multi_label:
        :param kwargs:
        """
        super(ClassificationModel, self).__init__(embedding, hyper_parameters, **kwargs)
        self.multi_label = multi_label
        self.multi_label_binarizer: MultiLabelBinarizer = None

        if self.multi_label:
            if not hyper_parameters or \
                    hyper_parameters.get('compile_params', {}).get('loss') is None:
                self.hyper_parameters['compile_params']['loss'] = 'binary_crossentropy'
            else:
                logging.warning('recommend to use binary_crossentropy loss for multi_label task')

            if not hyper_parameters or \
                    hyper_parameters.get('compile_params', {}).get('metrics') is None:
                self.hyper_parameters['compile_params']['metrics'] = ['categorical_accuracy']
            else:
                logging.warning('recommend to use categorical_accuracy metrivs for multi_label task')

            if not hyper_parameters or \
                    hyper_parameters.get('activation_layer', {}).get('sigmoid') is None:
                self.hyper_parameters['activation_layer']['activation'] = 'sigmoid'
            else:
                logging.warning('recommend to use sigmoid activation for multi_label task')

    def info(self):
        info = super(ClassificationModel, self).info()
        info['model_info']['multi_label'] = self.multi_label
        return info

    @property
    def label2idx(self) -> Dict[str, int]:
        return self._label2idx

    @property
    def token2idx(self) -> Dict[str, int]:
        return self.embedding.token2idx

    @label2idx.setter
    def label2idx(self, value):
        self._label2idx = value
        self._idx2label = dict([(val, key) for (key, val) in value.items()])

    def build_model(self,
                    x_train: List[List[str]],
                    y_train: Union[List[str], List[List[str]], List[Tuple[str]]],
                    x_validate: List[List[str]] = None,
                    y_validate: Union[List[str], List[List[str]], List[Tuple[str]]] = None):
        """
        build model function
        :return:
        """
        assert len(x_train) == len(y_train)
        self.build_token2id_label2id_dict(x_train, y_train, x_validate, y_validate)

        if not self.model:
            if self.embedding.sequence_length == 0:
                self.embedding.sequence_length = sorted([len(x) for x in x_train])[int(0.95 * len(x_train))]
                logging.info('sequence length set to {}'.format(self.embedding.sequence_length))
        self._prepare_model()
        self._compile_model()
        self.model.summary()

    @classmethod
    def load_model(cls, model_path: str):
        agent: ClassificationModel = super(ClassificationModel, cls).load_model(model_path)
        agent.multi_label = agent.model_info.get('multi_label', False)
        if agent.multi_label:
            keys = list(agent.label2idx.keys())
            agent.multi_label_binarizer = MultiLabelBinarizer(classes=keys)
            agent.multi_label_binarizer.fit(keys[0])
        return agent

    def build_token2id_label2id_dict(self,
                                     x_train: List[List[str]],
                                     y_train: List[str],
                                     x_validate: List[List[str]] = None,
                                     y_validate: List[str] = None):
        if x_validate:
            x_data = [*x_train, *x_validate]
            y_data = [*y_train, *y_validate]
        else:
            x_data = x_train
            y_data = y_train
        x_data_level = depth_count(x_data)
        if x_data_level > 2:
            for _ in range(x_data_level-2):
                x_data = list(chain(*x_data))

        self.embedding.build_token2idx_dict(x_data, 3)

        if self.multi_label:
            label_set = set()
            for i in y_data:
                label_set = label_set.union(list(i))
        else:
            label_set = set(y_data)

        if not len(self.label2idx):
            label2idx = {
                k.PAD: 0,
            }
            for idx, label in enumerate(label_set):
                label2idx[label] = idx + 1
            self._label2idx = label2idx
            self._idx2label = dict([(val, key) for (key, val) in label2idx.items()])
            self.multi_label_binarizer = MultiLabelBinarizer(classes=list(self.label2idx.keys()))


    def convert_label_to_idx(self, label: Union[List[str], str]) -> Union[List[int], int]:
        if isinstance(label, str):
            return self.label2idx[label]
        else:
            return [self.label2idx[l] for l in label]

    def convert_idx_to_label(self, token: Union[List[int], int]) -> Union[List[str], str]:
        if isinstance(token, int):
            return self._idx2label[token]
        else:
            return [self._idx2label[l] for l in token]

    def get_data_generator(self,
                           x_data: Union[List[List[str]], List[List[List[str]]]],
                           y_data: List[str],
                           batch_size: int = 64,
                           is_bert: bool = False):
        x_data_level = depth_count(x_data)
        if x_data_level == 2:
            x_data = [x_data]
        data_len = len(y_data)
        for x in x_data:
            assert len(x) == data_len
        while True:
            page_list = list(range((data_len // batch_size) + 1))
            random.shuffle(page_list)
            for page in page_list:
                start_index = page * batch_size
                end_index = start_index + batch_size
                target_x = []
                for x in x_data:
                    target_x.append(x[start_index: end_index])
                target_y = y_data[start_index: end_index]
                if len(target_x[0]) == 0:
                    target_x.pop()
                    for x in x_data:
                        target_x.append(x[0: batch_size])
                    target_y = y_data[0: batch_size]

                padded_x = []
                for i, x in enumerate(target_x):
                    tokenized_x = self.embedding.tokenize(x)

                    if isinstance(self.embedding.sequence_length, int):
                        padded_x.append(sequence.pad_sequences(tokenized_x,
                                                      maxlen=self.embedding.sequence_length,
                                                      padding='post')
                                )
                    elif isinstance(self.embedding.sequence_length, list):
                        padded_x.append(sequence.pad_sequences(tokenized_x,
                                                      maxlen=self.embedding.sequence_length[i],
                                                      padding='post')
                                )

                if self.multi_label:
                    padded_y = self.multi_label_binarizer.fit_transform(target_y)
                else:
                    tokenized_y = self.convert_label_to_idx(target_y)
                    padded_y = to_categorical(tokenized_y,
                                              num_classes=len(self.label2idx),
                                              dtype=np.int)
                if is_bert:
                    if isinstance(self.embedding.sequence_length, int):
                        padded_x_seg = [np.zeros(shape=(len(padded_x_i),
                                                        self.embedding.sequence_length))
                                                            for padded_x_i in padded_x]
                    elif isinstance(self.embedding.sequence_length, list):
                        padded_x_seg = [np.zeros(shape=(len(padded_x_i),
                                                        self.embedding.sequence_length[i]))
                                                            for i, padded_x_i in enumerate(padded_x)]
                    x_input_data = list(chain(*[(x, x_seg)
                                    for x, x_seg in zip(padded_x, padded_x_seg)]))
                else:
                    x_input_data = padded_x[0] if x_data_level == 2 else padded_x
                yield (x_input_data, padded_y)

    def fit(self,
            x_train: Union[List[List[str]], List[List[List[str]]]],
            y_train: Union[List[str], List[List[str]], List[Tuple[str]]],
            x_validate: Union[List[List[str]], List[List[List[str]]]] = None,
            y_validate: Union[List[str], List[List[str]], List[Tuple[str]]] = None,
            batch_size: int = 64,
            epochs: int = 5,
            class_weight: bool = False,
            fit_kwargs: Dict = None,
            **kwargs):
        """

        :param x_train: list of training data.
        :param y_train: list of training target label data.
        :param x_validate: list of validation data.
        :param y_validate: list of validation target label data.
        :param batch_size: batch size for trainer model
        :param epochs: Number of epochs to train the model.
        :param class_weight: set class weights for imbalanced classes
        :param fit_kwargs: additional kwargs to be passed to
               :func:`~keras.models.Model.fit`
        :param kwargs:
        :return:
        """
        x_train_level = depth_count(x_train)
        if x_train_level == 2:
            assert len(x_train) == len(y_train)
        elif x_train_level > 2:
            for x_part in x_train:
                assert len(x_part) == len(y_train)
        else:
            raise Exception('x_train type error')
       
        if len(y_train) < batch_size:
            batch_size = len(y_train) // 2

        if not self.model:
            if isinstance(self.embedding.sequence_length, int):
                if self.embedding.sequence_length == 0:
                    self.embedding.sequence_length = sorted([len(x) for x in x_train])[int(0.95 * len(x_train))]
                    logging.info('sequence length set to {}'.format(self.embedding.sequence_length))
            elif isinstance(self.embedding.sequence_length, list):
                seq_len = []
                for i, x_part in enumerate(x_train):
                    if self.embedding.sequence_length[i] == 0:
                        seq_len.append(max(sorted([len(x) for x in x_part])[int(0.95 * len(x_part))], 1))
                        logging.info(f'sequence_{i} length set to {self.embedding.sequence_length[i]}')
                    else:
                        seq_len.append(self.embedding.sequence_length[i])
                self.embedding.sequence_length = seq_len
            self.build_model(x_train, y_train, x_validate, y_validate)

        train_generator = self.get_data_generator(x_train,
                                                  y_train,
                                                  batch_size,
                                                  is_bert=self.embedding.is_bert)

        if fit_kwargs is None:
            fit_kwargs = {}

        if x_validate:
            validation_generator = self.get_data_generator(x_validate,
                                                           y_validate,
                                                           batch_size,
                                                           is_bert=self.embedding.is_bert)
            fit_kwargs['validation_data'] = validation_generator
            fit_kwargs['validation_steps'] = max(len(y_validate) // batch_size, 1)

        if class_weight:
            if self.multi_label:
                y_list = [self.convert_label_to_idx(y) for y in y_train]
                y_list = [y for ys in y_list for y in ys]
            else:
                y_list = self.convert_label_to_idx(y_train)
            class_weights = class_weight_calculte.compute_class_weight('balanced',
                                                                       np.unique(y_list),
                                                                       y_list)
        else:
            class_weights = None

        self.model.fit_generator(train_generator,
                                 steps_per_epoch=len(y_train) // batch_size,
                                 epochs=epochs,
                                 class_weight=class_weights,
                                 **fit_kwargs)

    def _format_output_dic(self, words: List[str], res: np.ndarray):
        results = sorted(list(enumerate(res)), key=lambda x: -x[1])
        candidates = []
        for result in results:
            candidates.append({
                'name': self.convert_idx_to_label([result[0]])[0],
                'confidence': float(result[1]),
            })

        data = {
            'words': words,
            'class': candidates[0],
            'class_candidates': candidates
        }
        return data

    def predict(self,
                sentence: Union[List[str], List[List[str]], List[List[List[str]]]],
                batch_size=None,
                output_dict=False,
                multi_label_threshold=0.6,
                debug_info=False) -> Union[List[str], str, List[Dict], Dict]:
        """
        predict with model
        :param sentence: single sentence as List[str] or list of sentence as List[List[str]]
        :param batch_size: predict batch_size
        :param output_dict: return dict with result with confidence
        :param multi_label_threshold:
        :param debug_info: print debug info using logging.debug when True
        :return:
        """
        sentence_level = depth_count(sentence)
        if sentence_level == 2:
            sentence = [sentence]
        elif sentence_level == 1:
            sentence = [[sentence]]
        padded_tokens = []
        for i, sent_part in enumerate(sentence):
            tokens = self.embedding.tokenize(sent_part)
            if isinstance(self.embedding.sequence_length, int):
                padded_tokens_part = sequence.pad_sequences(tokens,
                                                   maxlen=self.embedding.sequence_length,
                                                   padding='post')
                padded_tokens.append(padded_tokens_part)
                if self.embedding.is_bert:
                    padded_tokens.append(np.zeros(shape=(len(padded_tokens_part),
                                                         self.embedding.sequence_length)))
            elif isinstance(self.embedding.sequence_length, list):
                padded_tokens_part = sequence.pad_sequences(tokens,
                                                   maxlen=self.embedding.sequence_length[i],
                                                   padding='post')
                padded_tokens.append(padded_tokens_part)
                if self.embedding.is_bert:
                    padded_tokens.append(np.zeros(shape=(len(padded_tokens_part),
                                                         self.embedding.sequence_length[i])))
 
        x = padded_tokens
        res = self.model.predict(x, batch_size=batch_size)

        if self.multi_label:
            if debug_info:
                logging.info('raw output: {}'.format(res))
            res[res >= multi_label_threshold] = 1
            res[res < multi_label_threshold] = 0
            predict_result = res
        else:
            predict_result = res.argmax(-1)

        if debug_info:
            logging.info('input: {}'.format(x))
            logging.info('output: {}'.format(res))
            logging.info('output argmax: {}'.format(predict_result))

        if output_dict:
            words_list: List[List[str]] = sentence[0]
            results = []
            for index in range(len(words_list)):
                results.append(self._format_output_dic(words_list[index], res[index]))
            if sentence_level >= 2:
                return results
            elif sentence_level == 1:
                return results[0]
        else:
            if self.multi_label:
                results = self.multi_label_binarizer.inverse_transform(predict_result)
            else:
                results = self.convert_idx_to_label(predict_result)
            if sentence_level >= 2:
                return results
            elif sentence_level == 1:
                return results[0]

    def evaluate(self, x_data, y_data, batch_size=None, digits=4, debug_info=False) -> Tuple[float, float, Dict]:
        y_pred = self.predict(x_data, batch_size=batch_size)
        report = metrics.classification_report(y_data, y_pred, output_dict=True, digits=digits)
        print(metrics.classification_report(y_data, y_pred, digits=digits))
        if debug_info:
            for index in random.sample(list(range(len(x_data))), 5):
                logging.debug('------ sample {} ------'.format(index))
                logging.debug('x      : {}'.format(x_data[index]))
                logging.debug('y      : {}'.format(y_data[index]))
                logging.debug('y_pred : {}'.format(y_pred[index]))
        return report



# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: __version__.py
@time: 2019-02-21 15:22

"""

__version__ = '0.2.4'



# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: embedding
@time: 2019-01-20

"""
import json
import logging
import os
from typing import Dict, Any
from itertools import chain
from collections import Counter

import keras_bert
import numpy as np
from gensim.models import KeyedVectors
from keras.layers import Input, Embedding, concatenate
from keras.models import Model
from keras.preprocessing import sequence
from keras_gpt_2 import load_trained_model_from_checkpoint, get_bpe_from_files, BytePairEncoding

import kashgari.macros as k
from kashgari.type_hints import *
from kashgari.utils import helper
from kashgari.layers import NonMaskingLayer



EMBEDDINGS_PATH = os.path.join(k.DATA_PATH, 'embedding')


class BaseEmbedding(object):
    base_dict = {
        k.PAD: 0,
        k.BOS: 1,
        k.EOS: 2,
        k.UNK: 3
    }

    special_tokens = {
        k.PAD: k.PAD,
        k.UNK: k.UNK,
        k.BOS: k.BOS,
        k.EOS: k.EOS
    }

    def __init__(self,
                 name_or_path: str,
                 sequence_length: int = None,
                 embedding_size: int = None,
                 **kwargs):
        """
        init a WordEmbedding
        :param name_or_path: model name as `sgns.weibo.bigram` or model path like '/home/brikerman/w2v.model
        :param sequence_length: length of max sequence, all embedding is shaped as (sequence_length, embedding_size)
        :param embedding_size: embedding vector size, only need to set when using a CustomEmbedding
        :param kwargs: kwargs to pass to the method, func: `BaseEmbedding.build`
        """
        self.embedding_type = 'base'
        self.name = name_or_path
        self.embedding_size = embedding_size
        self._sequence_length = sequence_length
        self.model_path = ''
        self._token2idx: Dict[str, int] = None
        self._idx2token: Dict[int, str] = None
        self._model: Model = None
        self._kwargs = kwargs
        self.build(**kwargs)

    def update(self, info: Dict[str, Any]):
        self.name = info['name']
        self.embedding_type = info['embedding_type']
        self.embedding_size = info['embedding_size']
        self._sequence_length = info['sequence_length']
        self.model_path = info['model_path']
        self._kwargs = info['kwargs']

    def info(self):
        return {
            'embedding_type': self.embedding_type,
            'name': self.name,
            'embedding_size': self.embedding_size,
            'sequence_length': self._sequence_length,
            'model_path': self.model_path,
            'kwargs': self._kwargs
        }

    @property
    def token_count(self):
        return len(self._token2idx)

    @property
    def sequence_length(self):
        return self._sequence_length

    @sequence_length.setter
    def sequence_length(self, val):
        self._sequence_length = val
        self.build(**self._kwargs)

    @property
    def model(self) -> Model:
        return self._model

    @property
    def token2idx(self):
        return self._token2idx

    @property
    def is_bert(self):
        return self.embedding_type == 'bert'

    @token2idx.setter
    def token2idx(self, value):
        self._token2idx = value
        self._idx2token = dict([(value, key) for (key, value) in value.items()])

    @property
    def idx2token(self):
        return self._idx2token

    def build(self, **kwargs):
        raise NotImplementedError()

    def build_token2idx_dict(self, x_data: List[TextSeqType], min_count: int = 5):
        raise NotImplementedError()

    def tokenize(self,
                 sentence: TextSeqInputType,
                 add_bos_eos: bool = True) -> TokenSeqInputType:
        is_list = isinstance(sentence[0], list)

        def tokenize_sentence(text: TextSeqType) -> TokenSeqType:
            tokens = [self.token2idx.get(token, self.token2idx[k.UNK]) for token in text]
            if add_bos_eos:
                tokens = [self.token2idx[k.BOS]] + tokens + [self.token2idx[k.EOS]]
            return tokens

        if is_list:
            return [tokenize_sentence(sen) for sen in sentence]
        else:
            return tokenize_sentence(sentence)

    def embed(self, sentence: TextSeqInputType, seq_idx: int=0) -> np.array:
        is_list = isinstance(sentence[0], list)
        tokens = self.tokenize(sentence)

        if not is_list:
            tokens = [tokens]
        if isinstance(self.sequence_length, int):
            embed_input = sequence.pad_sequences(tokens, self.sequence_length, padding='post')
        elif isinstance(self.sequence_length, list):
            embed_input = sequence.pad_sequences(tokens, self.sequence_length[seq_idx], padding='post')

        embed_input = self.prepare_model_input(embed_input)
        print(embed_input)
        embed_pred = self.model.predict(embed_input)
        if is_list:
            return embed_pred
        else:
            return embed_pred[0]

    def prepare_model_input(self, input_x: np.array, **kwargs) -> np.array:
        return input_x


class WordEmbeddings(BaseEmbedding):
    base_dict = {
        k.PAD: 0,
        k.BOS: 1,
        k.EOS: 2,
        k.UNK: 3
    }

    URL_MAP = {
        'sgns.renmin.bigram': 'embedding/word2vec/sgns.renmin.bigram.bz2',
        'sgns.renmin.bigram-char': 'embedding/word2vec/sgns.renmin.bigram-char.bz2',
        'sgns.weibo.bigram': 'embedding/word2vec/sgns.weibo.bigram.bz2',
        'sgns.weibo.bigram-char': 'embedding/word2vec/sgns.weibo.bigram-char.bz2',
    }

    def get_embedding_matrix(self) -> np.array:
        base_matrix = []

        file = os.path.join(k.DATA_PATH, 'w2v_embedding_{}.json'.format(self.embedding_size))
        if os.path.exists(file):
            base_matrix = json.load(open(file, 'r', encoding='utf-8'))
            base_matrix = [np.array(matrix) for matrix in base_matrix]
        else:
            for index, key in enumerate(k.MARKED_KEYS):
                if index != 0:
                    vector = np.random.uniform(-0.5, 0.5, self.embedding_size)
                else:
                    vector = np.zeros(self.embedding_size)
                base_matrix.append(vector)
            with open(file, 'w', encoding='utf-8') as f:
                f.write(json.dumps([list(item) for item in base_matrix]))

        matrix_list = base_matrix + list(self.keyed_vector.vectors)
        return np.array(matrix_list)

    def build(self, **kwargs):
        self.embedding_type = 'word2vec'
        if self.name in WordEmbeddings.URL_MAP:
            url = self.URL_MAP.get(self.name)
            self.name = self.name + '.bz2'
        else:
            url = None

        self.model_path = helper.cached_path(self.name,
                                             url,
                                             sub_folders=['embedding', 'word2vec'])

        self.keyed_vector: KeyedVectors = KeyedVectors.load_word2vec_format(self.model_path, **kwargs)
        self.embedding_size = self.keyed_vector.vector_size

        word2idx = self.base_dict.copy()
        for word in self.keyed_vector.index2entity:
            word2idx[word] = len(word2idx)
        self.token2idx = word2idx

        input_layer = Input(shape=(self.sequence_length,), dtype='int32')
        embedding_matrix = self.get_embedding_matrix()

        current = Embedding(self.token_count,
                            self.embedding_size,
                            input_length=self.sequence_length,
                            weights=[embedding_matrix],
                            trainable=False)(input_layer)
        self._model = Model(input_layer, current)
        logging.debug('------------------------------------------------')
        logging.debug('Loaded gensim word2vec model')
        logging.debug('model        : {}'.format(self.model_path))
        logging.debug('word count   : {}'.format(len(self.keyed_vector.index2entity)))
        logging.debug('Top 50 word  : {}'.format(self.keyed_vector.index2entity[:50]))
        logging.debug('------------------------------------------------')

    def build_token2idx_dict(self, x_data: List[TextSeqType], min_count: int = 5):
        logging.debug("word2vec embedding no need to build token2idx with corpus")


class BERTEmbedding(BaseEmbedding):
    base_dict = {}
    special_tokens = {
        k.PAD: '[PAD]',
        k.UNK: '[UNK]',
        k.BOS: '[CLS]',
        k.EOS: '[SEP]',
    }

    model_key_map = {
        'bert-base-uncased': 'uncased_L-12_H-768_A-12',
        'bert-large-uncased': 'uncased_L-24_H-1024_A-16',
        'bert-base-cased': 'cased_L-12_H-768_A-12',
        'bert-large-cased': 'cased_L-24_H-1024_A-16',
        'bert-base-multilingual-cased': 'multi_cased_L-12_H-768_A-12',
        'bert-base-chinese': 'chinese_L-12_H-768_A-12'
    }

    pre_trained_models = {
        # BERT-Base, Uncased: 12-layer, 768-hidden, 12-heads, 110M parameters
        'uncased_L-12_H-768_A-12': 'https://storage.googleapis.com/bert_models/2018_10_18/'
                                   'uncased_L-12_H-768_A-12.zip',
        # BERT-Large, Uncased
        # 24-layer, 1024-hidden, 16-heads, 340M parameters
        'uncased_L-24_H-1024_A-16': 'https://storage.googleapis.com/bert_models/2018_10_18/'
                                    'uncased_L-24_H-1024_A-16.zip',
        # BERT-Base, Cased
        # 12-layer, 768-hidden, 12-heads , 110M parameters
        'cased_L-12_H-768_A-12': 'https://storage.googleapis.com/bert_models/2018_10_18/'
                                 'cased_L-12_H-768_A-12.zip',
        # BERT-Large, Cased
        # 24-layer, 1024-hidden, 16-heads, 340M parameters
        'cased_L-24_H-1024_A-16': 'https://storage.googleapis.com/bert_models/2018_10_18/'
                                  'cased_L-24_H-1024_A-16.zip',
        # BERT-Base, Multilingual Cased (New, recommended)
        # 104 languages, 12-layer, 768-hidden, 12-heads, 110M parameters
        'multi_cased_L-12_H-768_A-12': 'https://storage.googleapis.com/bert_models/2018_11_23/'
                                       'multi_cased_L-12_H-768_A-12.zip',
        # BERT-Base, Multilingual Uncased (Orig, not recommended)
        # 12-layer, 768-hidden, 12-heads, 110M parameters
        'multilingual_L-12_H-768_A-12': 'https://storage.googleapis.com/bert_models/2018_11_03/'
                                        'multilingual_L-12_H-768_A-12.zip',
        # BERT-Base, Chinese
        # Chinese Simplified and Traditional, 12-layer, 768-hidden, 12-heads, 110M
        'chinese_L-12_H-768_A-12': 'https://storage.googleapis.com/bert_models/2018_11_03/'
                                   'chinese_L-12_H-768_A-12.zip',
    }


    def __init__(self,
                 name_or_path: str,
                 sequence_length: int = None,
                 embedding_size: int = None,
                 layer_nums: int = 4,
                 trainable: bool = False,
                 **kwargs,):
        """
        init a WordEmbedding
        :param name_or_path: model name as `sgns.weibo.bigram` or model path like '/home/brikerman/w2v.model
        :param sequence_length: length of max sequence, all embedding is shaped as (sequence_length, embedding_size)
        :param embedding_size: embedding vector size, only need to set when using a CustomEmbedding
        :param layer_nums: number of layers whose outputs will be concatenated as a single output.
                           default `4`, the last 4 hidden layers
        :param trainable: whether if the output feature layer is trainable, default `False` and set it to `True` for finetune
        :param kwargs: kwargs to pass to the method, func: `BaseEmbedding.build`
        """
        self.layer_nums = layer_nums
        self.trainable = trainable
        self.training = False # We do not need to train the whole bert model so set it to `False`
        super(BERTEmbedding, self).__init__(name_or_path, sequence_length, embedding_size, **kwargs)


    def build(self):
        self.embedding_type = 'bert'
        url = self.pre_trained_models.get(self.model_key_map.get(self.name, self.name))
        self.model_path = helper.cached_path(self.model_key_map.get(self.name, self.name),
                                             url,
                                             ['embedding', 'bert'])

        config_path = os.path.join(self.model_path, 'bert_config.json')
        check_point_path = os.path.join(self.model_path, 'bert_model.ckpt')
        logging.info('loading bert model from {}\n'.format(self.model_path))
        model = keras_bert.load_trained_model_from_checkpoint(config_path,
                                                              check_point_path,
                                                              seq_len=self.sequence_length,
                                                              output_layer_num=self.layer_nums,
                                                              training=self.training,
                                                              trainable=self.trainable
                                                              )
        #num_layers = len(model.layers)
        #features_layers = [model.get_layer(index=num_layers-1+idx*8).output\
        #                    for idx in range(-3, 1)]
        #embedding_layer = concatenate(features_layers)
        #output_layer = NonMaskingLayer()(embedding_layer)
        output_layer = NonMaskingLayer()(model.output)
        self._model = Model(model.inputs, output_layer)

        self.embedding_size = self.model.output_shape[-1]
        dict_path = os.path.join(self.model_path, 'vocab.txt')
        word2idx = {}
        with open(dict_path, 'r', encoding='utf-8') as f:
            words = f.read().splitlines()
        for word in words:
            word2idx[word] = len(word2idx)
        for key, value in self.special_tokens.items():
            word2idx[key] = word2idx[value]

        self.token2idx = word2idx

    def build_token2idx_dict(self, x_data: List[TextSeqType], min_count: int = 5):
        logging.debug("bert embedding no need to build token2idx with corpus")

    def prepare_model_input(self, input_x: np.array, **kwargs) -> np.array:
        input_seg = np.zeros(input_x.shape)
        return [input_x, input_seg]


class CustomEmbedding(BaseEmbedding):
    def __init__(self,
                 name_or_path: str = 'custom-embedding',
                 sequence_length: int = None,
                 embedding_size: int = None,
                 **kwargs):
        """
        :param name_or_path: just a name for custom embedding
        :param sequence_length: length of max sequence, all embedding is shaped as (sequence_length, embedding_size)
        :param embedding_size: embedding vector size, only need to set when using a CustomEmbedding
        :param kwargs: kwargs to pass to the method, func: `BaseEmbedding.build`
        """
        if sequence_length is None or embedding_size is None:
            raise ValueError('Must set sequence_length and sequence_length when using the CustomEmbedding layer')
        super(CustomEmbedding, self).__init__(name_or_path, sequence_length, embedding_size, **kwargs)

    def build(self, **kwargs):
        if self._token2idx is None:
            logging.debug('need to build after build_word2idx')
        else:
            input_x = Input(shape=(self.sequence_length,), dtype='int32')
            current = Embedding(self.token_count,
                                self.embedding_size)(input_x)
            self._model = Model(input_x, current)

    def build_token2idx_dict(self, x_data: List[TextSeqType], min_count: int = 5):
        if self.token2idx is None:
            #word_set: Dict[str, int] = {}
            # for x_item in x_data:
            #     for word in x_item:
            #         word_set[word] = word_set.get(word, 0) + 1
            data_depth = helper.depth_count(x_data)
            if data_depth > 1:
                x_items = x_data
                for _ in range(data_depth-1):
                    x_items = list(chain(*x_items))
            word_freq = Counter(x_items)
            # word_set = {word: freq for word, freq in word_freq.items() if freq >= min_count}
            # word2idx_list = sorted(word_set.items(), key=lambda kv: -kv[1])
            word2idx_list = sorted(word_freq.items(), key=lambda kv: -kv[1])

            word2idx = self.base_dict.copy()
            offset = len(word2idx)
            # for word, count in word2idx_list:
            #     if count >= min_count:
            #         word2idx[word] = len(word2idx)
            for idx, (word, freq) in enumerate(word2idx_list):
                if freq >= min_count:
                    word2idx[word] = idx + offset

            self.token2idx = word2idx
        self.build()


class TwoHeadEmbedding(CustomEmbedding):
    def __init__(self,
                 name_or_path: str = 'twohead-embedding',
                 sequence_length: List[int] = None,
                 embedding_size: int = None,
                 **kwargs):
        """
        Inheritated from CustomEmbedding class.
        :param name_or_path: just a name for two head embedding
        :param sequence_length: max length list of sequences, all embedding is shaped as (sequence_length[idx], embedding_size)
        :param embedding_size: embedding vector size, only need to set when using a CustomEmbedding or its subclass
        :param kwargs: kwargs to pass to the method, func: `BaseEmbedding.build`
        """
        if sequence_length is None or embedding_size is None:
            raise ValueError('Must set all sequence_length and embedding_size when using the TwoheadEmbedding layer')
        super(TwoHeadEmbedding, self).__init__(name_or_path, sequence_length, embedding_size, **kwargs)

    def build(self, **kwargs):
        self.embedding_type = 'twohead'
        if self._token2idx is None:
            logging.debug('need to build after build_word2idx')
        else:
            input_x1 = Input(shape=(self.sequence_length[0],), dtype='int32', name='master_input')
            current1 = Embedding(self.token_count,
                                self.embedding_size)(input_x1)
            input_x2 = Input(shape=(self.sequence_length[1],), dtype='int32', name='assist_input')
            current2 = Embedding(self.token_count,
                                self.embedding_size)(input_x2)
            current = concatenate([current1, current2], axis=1)
            self._model = Model(inputs=[input_x1, input_x2], outputs=current)

    def build_token2idx_dict(self, x_data: List[TextSeqType], min_count: int = 5):
        super(TwoHeadEmbedding, self).build_token2idx_dict(x_data, min_count)

    def embed(self, sentences_pair: List[List[TextSeqInputType]]) -> np.array:
        embed_inputs = []
        for idx, sentences in enumerate(sentences_pair):
            is_list = isinstance(sentences[0], list)
            tokens = self.tokenize(sentences)
            if not is_list:
                tokens = [tokens]
            if isinstance(self.sequence_length, list):
                embed_input = sequence.pad_sequences(tokens, self.sequence_length[idx], padding='post')
            elif isinstance(self.sequence_length, int):
                embed_input = sequence.pad_sequences(tokens, self.sequence_length, padding='post')
            embed_inputs.append(embed_input)
        embed_inputs = self.prepare_model_input(embed_inputs)
        print(embed_inputs)
        embed_pred = self.model.predict(embed_inputs)
        return embed_pred


class GPT2Embedding(BaseEmbedding):

    def build(self, **kwargs):
        self.embedding_type = 'gpt2'

        config_path = os.path.join(self.name, 'hparams.json')
        checkpoint_path = os.path.join(self.name, 'model.ckpt')
        encoder_path = os.path.join(self.name, 'encoder.json')
        vocab_path = os.path.join(self.name, 'vocab.bpe')

        self._model: Model = load_trained_model_from_checkpoint(config_path, checkpoint_path)
        for layer in self._model.layers:
            layer.trainable = False

        self.bpe: BytePairEncoding = get_bpe_from_files(encoder_path, vocab_path)

        word2idx = self.bpe.token_dict.copy()
        word2idx[k.PAD] = word2idx['pad']
        word2idx[k.UNK] = word2idx['unk']
        word2idx[k.BOS] = word2idx['pad']
        word2idx[k.EOS] = word2idx['pad']
        self.token2idx = word2idx

    def build_token2idx_dict(self, x_data: List[TextSeqType], min_count: int = 5):
        logging.debug("word2vec embedding no need to build token2idx with corpus")


if __name__ == '__main__':
    train_x = [
        list('语言学（英语：linguistics）是一门关于人类语言的科学研究'),
        list('语言学（英语：linguistics）是一门关于人类语言的科学研究'),
        list('语言学（英语：linguistics）是一门关于人类语言的科学研究'),
        list('语言学包含了几种分支领域。'),
        list('在语言结构（语法）研究与意义（语义与语用）研究之间存在一个重要的主题划分'),
    ]
    train_y = ['a', 'a', 'a', 'b', 'c']

    from kashgari.utils.logger import init_logger
    from kashgari.tasks.classification import CNNModel
    init_logger()
    embedding = GPT2Embedding('/Users/brikerman/Desktop/python/gpt-2/models/117M', 10)
    r = embedding.embed(['hello', 'world'])
    model = CNNModel(embedding)
    model.fit(train_x, train_y, epochs=20)
    print(r.shape)
